#
# Copyright (C) 2024 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import configparser
import logging
from typing import List, Union


class InvalidConfigError(Exception):
    pass


DEFAULT_CONFIG = {
    'reader0': {
        'type': "lge450",
        'port': "/dev/ttyUSB0",
        'baudrate': 2400,
        'key': "",
    },
    'sink0': {
        'type': "logger",
        'name': "DataLogger",
    },
    'sink1': {
        'type': "mqtt",
        'host': "localhost",
        'port': 1883,
        'tls': False,
        'ca_file_path': "",
        'check_hostname': True,
        'username': "",
        'password': "",
        'client_cert_path': "",
        'client_key_path': "",
    },
    'logging': {
        'default': 'DEBUG',
        'collector': 'DEBUG',
        'smartmeter': 'DEBUG',
        'sink': 'DEBUG',
    }
}


def read_config_files(config_paths: Union[str, List[str]]) -> configparser.ConfigParser:
    config = configparser.ConfigParser()
    read_files = config.read(config_paths)
    if read_files:
        logging.info("Read following config files: %s", read_files)
    else:
        logging.warning("No config file found. Using default config.")
        config = get_default_config()
    return config


def get_default_config() -> configparser.ConfigParser:
    config = configparser.ConfigParser()
    config.read_dict(DEFAULT_CONFIG)
    return config


def write_default_config(config_path: str):
    config = get_default_config()
    with open(config_path, 'w', encoding='utf-8') as file:
        config.write(file, True)



#
# Copyright (C) 2024 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import configparser
import json
import sys
from datetime import datetime, timezone
from unittest import mock

import pytest
from asyncio_mqtt.error import MqttCodeError
from paho.mqtt.client import MQTT_ERR_NO_CONN
from pytest_mock.plugin import MockerFixture

from smartmeter_datacollector.sinks.mqtt_sink import MqttConfig, MqttDataSink
from smartmeter_datacollector.smartmeter.meter_data import MeterDataPoint, MeterDataPointType

TEST_TYPE = MeterDataPointType("TEST_TYPE", "test type", "unit")


@pytest.mark.asyncio
async def test_mqtt_sink_start_stop(mocker: MockerFixture):
    config = MqttConfig("localhost")
    sink = MqttDataSink(config)
    client_mock = mocker.patch.object(sink, "_client", autospec=True)

    await sink.start()
    client_mock.connect.assert_awaited_once()

    await sink.stop()
    client_mock.disconnect.assert_awaited_once()


@pytest.mark.asyncio
async def test_mqtt_sink_send_point_when_started(mocker: MockerFixture):
    config = MqttConfig("localhost")
    sink = MqttDataSink(config)
    client_mock = mocker.patch.object(sink, "_client", autospec=True)
    data_point = MeterDataPoint(TEST_TYPE, 1.0, "test_source", datetime.now(timezone.utc))
    expected_topic = f"smartmeter/test_source/{TEST_TYPE.identifier}"
    expected_payload = json.dumps({
        "value": data_point.value,
        "timestamp": int(data_point.timestamp.timestamp()),
    })

    await sink.start()
    await sink.send(data_point)

    client_mock.publish.assert_awaited_once_with(expected_topic, expected_payload)


@pytest.mark.asyncio
async def test_mqtt_sink_send_reconnect_when_not_started(mocker: MockerFixture):
    config = MqttConfig("localhost")
    sink = MqttDataSink(config)
    client_mock = mocker.patch.object(sink, "_client", autospec=True)
    data_point = MeterDataPoint(TEST_TYPE, 1.0, "test_source", datetime.now(timezone.utc))

    client_mock.publish.side_effect = MqttCodeError(MQTT_ERR_NO_CONN)
    await sink.send(data_point)

    client_mock.publish.assert_awaited()
    assert client_mock.publish.await_count == 2
    client_mock.connect.assert_awaited_once()


def test_mqtt_config_unencrypted_unauthorized():
    cfg_parser = configparser.ConfigParser()
    cfg_parser.read_dict({
        "sink": {
            'type': "mqtt",
            'host': "localhost",
            'port': 1883,
        }
    })
    sink_section = cfg_parser["sink"]
    cfg = MqttConfig.from_sink_config(sink_section)

    assert cfg.broker_host == sink_section.get("host")
    assert cfg.port == sink_section.getint("port")
    assert cfg.use_tls == False
    assert cfg.username is None
    assert cfg.password is None
    assert cfg.client_cert_path is None
    assert cfg.client_key_path is None

    sink = MqttDataSink(cfg)


def test_mqtt_config_encrypted_unauthorized():
    cfg_parser = configparser.ConfigParser()
    cfg_parser.read_dict({
        "sink": {
            'type': "mqtt",
            'host': "http://secure-host.com",
            'port': 8883,
            'tls': True,
            'ca_file_path': "/test/path",
            'check_hostname': False,
        }
    })
    sink_section = cfg_parser["sink"]
    cfg = MqttConfig.from_sink_config(sink_section)

    assert cfg.broker_host == sink_section.get("host")
    assert cfg.port == sink_section.getint("port")
    assert cfg.use_tls == True
    assert cfg.ca_cert_path == sink_section.get("ca_file_path")
    assert cfg.check_hostname == False
    assert cfg.username is None

    with mock.patch("smartmeter_datacollector.sinks.mqtt_sink.ssl"):
        sink = MqttDataSink(cfg)


def test_mqtt_config_encrypted_authorized_user_pass():
    cfg_parser = configparser.ConfigParser()
    cfg_parser.read_dict({
        "sink": {
            'type': "mqtt",
            'host': "http://secure-host.com",
            'port': 8883,
            'tls': True,
            'ca_file_path': "",
            'check_hostname': True,
            'username': "testuser",
            'password': "testpassword",
            'client_cert_path': "",
        }
    })
    sink_section = cfg_parser["sink"]
    cfg = MqttConfig.from_sink_config(sink_section)

    assert cfg.use_tls == True
    assert cfg.ca_cert_path == sink_section.get("ca_file_path")
    assert cfg.check_hostname == True
    assert cfg.username == sink_section.get("username")
    assert cfg.password == sink_section.get("password")
    assert cfg.client_cert_path is None
    assert cfg.client_key_path is None

    sink = MqttDataSink(cfg)


def test_mqtt_config_encrypted_authorized_client_cert():
    cfg_parser = configparser.ConfigParser()
    cfg_parser.read_dict({
        "sink": {
            'type': "mqtt",
            'host': "http://secure-host.com",
            'port': 8883,
            'tls': True,
            'ca_file_path': "/test/path",
            'check_hostname': True,
            'username': "",
            'client_cert_path': "/path/to/cert",
            'client_key_path': "rel/path/to/key",
        }
    })
    sink_section = cfg_parser["sink"]
    cfg = MqttConfig.from_sink_config(sink_section)

    assert cfg.use_tls == True
    assert cfg.ca_cert_path == sink_section.get("ca_file_path")
    assert cfg.check_hostname == True
    assert cfg.username is None
    assert cfg.password is None
    assert cfg.client_cert_path == sink_section.get("client_cert_path")
    assert cfg.client_key_path == sink_section.get("client_key_path")

    with mock.patch("smartmeter_datacollector.sinks.mqtt_sink.ssl"):
        sink = MqttDataSink(cfg)



#
# Copyright (C) 2024 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
from configparser import ConfigParser
from pathlib import Path

from smartmeter_datacollector import config


def test_default_config_has_sections():
    def_conf = config.get_default_config()
    assert isinstance(def_conf, ConfigParser)
    assert def_conf.has_section("reader0")
    assert def_conf.get("reader0", "type") == "lge450"

    assert def_conf.has_section("sink0")
    assert def_conf.get("sink0", "type") == "logger"
    assert def_conf.has_section("sink1")
    assert def_conf.get("sink1", "type") == "mqtt"

    assert def_conf.has_section("logging")


def test_write_default_config(tmp_path: Path):
    file_path = tmp_path / "cfg.ini"
    config.write_default_config(str(file_path))

    assert file_path.is_file()
    assert len(file_path.read_text(encoding="utf-8")) > 0


def test_read_config(tmp_path: Path):
    file_path = tmp_path / "cfg.ini"
    CFG = """
    [reader0]
    type = lge450
    port = /test
    key = 
    """
    file_path.write_text(CFG, encoding="utf-8")

    cfg = config.read_config_files(str(file_path))
    assert isinstance(cfg, ConfigParser)

    assert cfg.has_section("reader0")
    assert cfg.get("reader0", "type") == "lge450"
    assert cfg.get("reader0", "port") == "/test"
    assert cfg.get("reader0", "key") == ""



#
# Copyright (C) 2024 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import logging
import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Optional

from gurux_dlms import GXDateTime
from gurux_dlms.objects import GXDLMSClock, GXDLMSData, GXDLMSRegister

from .meter_data import MeterDataPointType, MeterDataPointTypes
from .obis import OBISCode

LOGGER = logging.getLogger("smartmeter")


@dataclass
class RegisterCosem:
    obis: OBISCode
    data_point_type: MeterDataPointType
    scaling: float = 1.0


DEFAULT_REGISTER_MAPPING = [
    RegisterCosem(OBISCode(1, 0, 1, 7, 0), MeterDataPointTypes.ACTIVE_POWER_P.value),
    RegisterCosem(OBISCode(1, 0, 2, 7, 0), MeterDataPointTypes.ACTIVE_POWER_N.value),
    RegisterCosem(OBISCode(1, 0, 3, 7, 0), MeterDataPointTypes.REACTIVE_POWER_P.value),
    RegisterCosem(OBISCode(1, 0, 4, 7, 0), MeterDataPointTypes.REACTIVE_POWER_N.value),

    RegisterCosem(OBISCode(1, 0, 13, 7, 0), MeterDataPointTypes.POWER_FACTOR.value, 0.001),
    RegisterCosem(OBISCode(1, 0, 14, 7, 0), MeterDataPointTypes.NET_FREQUENCY.value),

    RegisterCosem(OBISCode(1, 0, 21, 7, 0), MeterDataPointTypes.ACTIVE_POWER_P_L1.value),
    RegisterCosem(OBISCode(1, 0, 22, 7, 0), MeterDataPointTypes.ACTIVE_POWER_N_L1.value),
    RegisterCosem(OBISCode(1, 0, 23, 7, 0), MeterDataPointTypes.REACTIVE_POWER_P_L1.value),
    RegisterCosem(OBISCode(1, 0, 24, 7, 0), MeterDataPointTypes.REACTIVE_POWER_N_L1.value),

    RegisterCosem(OBISCode(1, 0, 31, 7, 0), MeterDataPointTypes.CURRENT_L1.value, 0.01),
    RegisterCosem(OBISCode(1, 0, 32, 7, 0), MeterDataPointTypes.VOLTAGE_L1.value),

    RegisterCosem(OBISCode(1, 0, 41, 7, 0), MeterDataPointTypes.ACTIVE_POWER_P_L2.value),
    RegisterCosem(OBISCode(1, 0, 42, 7, 0), MeterDataPointTypes.ACTIVE_POWER_N_L2.value),
    RegisterCosem(OBISCode(1, 0, 43, 7, 0), MeterDataPointTypes.REACTIVE_POWER_P_L2.value),
    RegisterCosem(OBISCode(1, 0, 44, 7, 0), MeterDataPointTypes.REACTIVE_POWER_N_L2.value),

    RegisterCosem(OBISCode(1, 0, 51, 7, 0), MeterDataPointTypes.CURRENT_L2.value, 0.01),
    RegisterCosem(OBISCode(1, 0, 52, 7, 0), MeterDataPointTypes.VOLTAGE_L2.value),

    RegisterCosem(OBISCode(1, 0, 61, 7, 0), MeterDataPointTypes.ACTIVE_POWER_P_L3.value),
    RegisterCosem(OBISCode(1, 0, 62, 7, 0), MeterDataPointTypes.ACTIVE_POWER_N_L3.value),
    RegisterCosem(OBISCode(1, 0, 63, 7, 0), MeterDataPointTypes.REACTIVE_POWER_P_L3.value),
    RegisterCosem(OBISCode(1, 0, 64, 7, 0), MeterDataPointTypes.REACTIVE_POWER_N_L3.value),

    RegisterCosem(OBISCode(1, 0, 71, 7, 0), MeterDataPointTypes.CURRENT_L3.value, 0.01),
    RegisterCosem(OBISCode(1, 0, 72, 7, 0), MeterDataPointTypes.VOLTAGE_L3.value),

    RegisterCosem(OBISCode(1, 0, 81, 7, 40), MeterDataPointTypes.ANGLE_UI_L1.value),
    RegisterCosem(OBISCode(1, 0, 81, 7, 51), MeterDataPointTypes.ANGLE_UI_L2.value),
    RegisterCosem(OBISCode(1, 0, 81, 7, 62), MeterDataPointTypes.ANGLE_UI_L3.value),

    RegisterCosem(OBISCode(1, 0, 1, 8, 0), MeterDataPointTypes.ACTIVE_ENERGY_P.value),
    RegisterCosem(OBISCode(1, 0, 1, 8, 1), MeterDataPointTypes.ACTIVE_ENERGY_P_T1.value),
    RegisterCosem(OBISCode(1, 0, 1, 8, 2), MeterDataPointTypes.ACTIVE_ENERGY_P_T2.value),
    RegisterCosem(OBISCode(1, 0, 2, 8, 0), MeterDataPointTypes.ACTIVE_ENERGY_N.value),
    RegisterCosem(OBISCode(1, 0, 2, 8, 1), MeterDataPointTypes.ACTIVE_ENERGY_N_T1.value),
    RegisterCosem(OBISCode(1, 0, 2, 8, 2), MeterDataPointTypes.ACTIVE_ENERGY_N_T2.value),
    RegisterCosem(OBISCode(1, 0, 3, 8, 0), MeterDataPointTypes.REACTIVE_ENERGY_P.value),
    RegisterCosem(OBISCode(1, 0, 3, 8, 1), MeterDataPointTypes.REACTIVE_ENERGY_P_T1.value),
    RegisterCosem(OBISCode(1, 0, 3, 8, 2), MeterDataPointTypes.REACTIVE_ENERGY_P_T2.value),
    RegisterCosem(OBISCode(1, 0, 4, 8, 0), MeterDataPointTypes.REACTIVE_ENERGY_N.value),
    RegisterCosem(OBISCode(1, 0, 4, 8, 1), MeterDataPointTypes.REACTIVE_ENERGY_N_T1.value),
    RegisterCosem(OBISCode(1, 0, 4, 8, 2), MeterDataPointTypes.REACTIVE_ENERGY_N_T2.value),

    RegisterCosem(OBISCode(1, 0, 5, 8, 0), MeterDataPointTypes.REACTIVE_ENERGY_Q1.value),
    RegisterCosem(OBISCode(1, 0, 5, 8, 1), MeterDataPointTypes.REACTIVE_ENERGY_Q1_T1.value),
    RegisterCosem(OBISCode(1, 0, 5, 8, 2), MeterDataPointTypes.REACTIVE_ENERGY_Q1_T2.value),

    RegisterCosem(OBISCode(1, 0, 6, 8, 0), MeterDataPointTypes.REACTIVE_ENERGY_Q2.value),
    RegisterCosem(OBISCode(1, 0, 6, 8, 1), MeterDataPointTypes.REACTIVE_ENERGY_Q2_T1.value),
    RegisterCosem(OBISCode(1, 0, 6, 8, 2), MeterDataPointTypes.REACTIVE_ENERGY_Q2_T2.value),

    RegisterCosem(OBISCode(1, 0, 7, 8, 0), MeterDataPointTypes.REACTIVE_ENERGY_Q3.value),
    RegisterCosem(OBISCode(1, 0, 7, 8, 1), MeterDataPointTypes.REACTIVE_ENERGY_Q3_T1.value),
    RegisterCosem(OBISCode(1, 0, 7, 8, 2), MeterDataPointTypes.REACTIVE_ENERGY_Q3_T2.value),

    RegisterCosem(OBISCode(1, 0, 8, 8, 0), MeterDataPointTypes.REACTIVE_ENERGY_Q4.value),
    RegisterCosem(OBISCode(1, 0, 8, 8, 1), MeterDataPointTypes.REACTIVE_ENERGY_Q4_T1.value),
    RegisterCosem(OBISCode(1, 0, 8, 8, 2), MeterDataPointTypes.REACTIVE_ENERGY_Q4_T2.value),
]


class Cosem:
    CLOCK_DEFAULT_OBIS = OBISCode(0, 0, 1, 0, 0)
    OBJECT_DETECT_ATTEMPTS = 3

    def __init__(self,
                 fallback_id: str,
                 id_obis_override: Optional[List[OBISCode]] = None,
                 register_obis_extended: Optional[List[RegisterCosem]] = None) -> None:
        self._id: Optional[str] = None
        if not fallback_id:
            fallback_id = str(uuid.uuid1())
            LOGGER.warning("Empty fallback ID. Setting to random UUID %s.", fallback_id)
        self._fallback_id = fallback_id
        self._id_obis_override = id_obis_override if id_obis_override else []
        registers = DEFAULT_REGISTER_MAPPING
        if register_obis_extended:
            registers += register_obis_extended
        self._register_obis = {r.obis: r for r in registers}
        self._id_detect_countdown = Cosem.OBJECT_DETECT_ATTEMPTS

    def retrieve_id(self, dlms_objects: Dict[OBISCode, Any]) -> str:
        if self._id:
            return self._id

        id_obis = self._find_obis_of_id(dlms_objects)
        if not id_obis:
            LOGGER.debug("Unable to find ID object. Using fallback ID %s.", self._fallback_id)
            self._trigger_id_detect_counter()
            return self._fallback_id

        id_obj = dlms_objects[id_obis]
        if not isinstance(id_obj, (GXDLMSData, GXDLMSRegister)):
            LOGGER.debug("Invalid ID object for OBIS code %s. Using fallback ID %s.", id_obis, self._fallback_id)
            self._trigger_id_detect_counter()
            return self._fallback_id

        meter_id = str(id_obj.getValues()[1])
        if len(meter_id) == 0:
            LOGGER.debug("Invalid ID for OBIS code %s. Using fallback ID %s.", id_obis, self._fallback_id)
            self._trigger_id_detect_counter()
            return self._fallback_id

        self._id = meter_id
        LOGGER.debug("ID %s found with OBIS code %s.", meter_id, id_obis)
        return meter_id

    def retrieve_time_from_dlms_registers(self, dlms_objects: Dict[OBISCode, Any]) -> Optional[datetime]:
        clock_obj = dlms_objects.get(Cosem.CLOCK_DEFAULT_OBIS, None)
        if clock_obj and isinstance(clock_obj, GXDLMSClock):
            timestamp = self._extract_datetime(clock_obj)
            if timestamp:
                # assume local time if tzinfo is None
                return timestamp.astimezone(timestamp.tzinfo)
        return None

    def get_register(self, obis: OBISCode) -> Optional[RegisterCosem]:
        return self._register_obis.get(obis, None)

    def _trigger_id_detect_counter(self):
        self._id_detect_countdown -= 1
        if self._id_detect_countdown <= 0:
            LOGGER.warning("Unable to retrieve ID after %i attempts. Using fallback ID %s from now on.",
                           Cosem.OBJECT_DETECT_ATTEMPTS, self._fallback_id)
            self._id = self._fallback_id

    def _find_obis_of_id(self, dlms_objects: Dict[OBISCode, Any]) -> Optional[OBISCode]:
        # first check all ID OBIS of override list (empty by default)
        for id_obis in self._id_obis_override:
            if id_obis in dlms_objects:
                return id_obis

        # then check for default ID OBIS
        for obj_obis in dlms_objects:
            if self._is_default_id_obis(obj_obis):
                return obj_obis

        return None

    @staticmethod
    def _is_default_id_obis(obis: OBISCode) -> bool:
        return (
            obis == OBISCode(0, 0, 42, 0, 0) or
            (obis.a == 0 and obis.c == 96 and obis.d == 1 and obis.e >= 0 and obis.e <= 9) or
            (obis.a == 1 and obis.c == 0 and obis.d == 0 and obis.e >= 0 and obis.e <= 9))

    @staticmethod
    def _extract_datetime(clock_object: GXDLMSClock) -> Optional[datetime]:
        time_obj: GXDateTime = clock_object.getValues()[1]
        if isinstance(time_obj, GXDateTime):
            return time_obj.value
        return None



#
# Copyright (C) 2024 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import logging
from datetime import datetime, timezone
from typing import Any, List, Optional, Tuple

from gurux_dlms import GXByteBuffer, GXDateTime, GXDLMSClient, GXReplyData
from gurux_dlms.enums import InterfaceType, ObjectType, Security
from gurux_dlms.objects import (GXDLMSCaptureObject, GXDLMSClock, GXDLMSData, GXDLMSObject, GXDLMSPushSetup,
                                GXDLMSRegister)
from gurux_dlms.secure import GXDLMSSecureClient

from .cosem import Cosem
from .meter_data import MeterDataPoint
from .obis import OBISCode

LOGGER = logging.getLogger("smartmeter")


class HdlcDlmsParser:
    HDLC_BUFFER_MAX_SIZE = 5000

    def __init__(self, cosem: Cosem, block_cipher_key: Optional[str] = None, use_system_time: bool = False) -> None:
        if block_cipher_key:
            self._client = GXDLMSSecureClient(
                useLogicalNameReferencing=True,
                interfaceType=InterfaceType.HDLC)
            self._client.ciphering.security = Security.ENCRYPTION
            self._client.ciphering.blockCipherKey = GXByteBuffer.hexToBytes(block_cipher_key)
        else:
            self._client = GXDLMSClient(
                useLogicalNameReferencing=True,
                interfaceType=InterfaceType.HDLC)

        self._hdlc_buffer = GXByteBuffer()
        self._dlms_data = GXReplyData()
        self._cosem = cosem
        self._use_system_time = use_system_time
        if use_system_time:
            LOGGER.info("Use system UTC time instead of time in DLMS messages for this smart meter.")

    def append_to_hdlc_buffer(self, data: bytes) -> None:
        if self._hdlc_buffer.getSize() > self.HDLC_BUFFER_MAX_SIZE:
            LOGGER.warning("HDLC byte-buffer > %i. Buffer is cleared, some data is lost.",
                           self.HDLC_BUFFER_MAX_SIZE)
            self._hdlc_buffer.clear()
            self._dlms_data.clear()
        self._hdlc_buffer.set(data)

    def clear_hdlc_buffer(self) -> None:
        self._hdlc_buffer.clear()

    def extract_data_from_hdlc_frames(self) -> bool:
        """
        Try to extract data fragments from HDLC frame-buffer and store it into DLMS buffer.
        HDLC buffer is being cleared.
        Returns: True if data is complete for parsing.
        """
        tmp = GXReplyData()
        try:
            LOGGER.debug("HDLC Buffer: %s", GXByteBuffer.hex(self._hdlc_buffer))
            self._client.getData(self._hdlc_buffer, tmp, self._dlms_data)
        except (ValueError, TypeError) as ex:
            LOGGER.warning("Failed to extract data from HDLC frame: '%s' Some data got lost.", ex)
            self._hdlc_buffer.clear()
            self._dlms_data.clear()
            return False

        if not self._dlms_data.isComplete():
            LOGGER.debug("HDLC frame incomplete and will not be parsed yet.")
            return False

        if self._dlms_data.isMoreData():
            LOGGER.debug("More DLMS data expected. Not yet ready to be parsed.")
            return False

        LOGGER.debug("DLMS packet complete and ready for parsing.")
        self._hdlc_buffer.clear()
        return True

    def extract_message_time(self) -> Optional[datetime]:
        if not isinstance(self._dlms_data.time, GXDateTime):
            return None
        dt = self._dlms_data.time.value
        if isinstance(dt, datetime):
            # assume local time if tzinfo is None
            return dt.astimezone(dt.tzinfo)
        return None

    def parse_to_dlms_objects(self) -> List[GXDLMSObject]:
        if not isinstance(self._dlms_data.value, list) or not self._dlms_data.value:
            self._dlms_data.clear()
            LOGGER.error("DLMS data is no list or empty list. Not parsable.")
            return []

        # pylint: disable=unsubscriptable-object
        if isinstance(self._dlms_data.value[0], list):
            # message with included push-object-list as first object
            dlms_objects = self._parse_dlms_with_push_object_list()
        else:
            # message without push-object-list
            dlms_objects = self._parse_dlms_without_push_list()
            if not dlms_objects:
                # message contains no OBIS codes, only values, which is not supported.
                LOGGER.warning("DLMS message is formatted in unknown structure and cannot be parsed by this software.")
        self._dlms_data.clear()
        return dlms_objects

    def convert_dlms_bundle_to_reader_data(self, dlms_objects: List[GXDLMSObject],
                                           message_time: Optional[datetime] = None) -> List[MeterDataPoint]:
        obis_obj_pairs = {}
        for obj in dlms_objects:
            try:
                obis_obj_pairs[OBISCode.from_string(str(obj.logicalName))] = obj
            except ValueError as ex:
                LOGGER.warning("Skipping unparsable DLMS object. (Reason: %s)", ex)

        meter_id = self._cosem.retrieve_id(obis_obj_pairs)

        timestamp = None
        if self._use_system_time:
            timestamp = datetime.now(timezone.utc)

        if not timestamp:
            timestamp = self._cosem.retrieve_time_from_dlms_registers(obis_obj_pairs)
        if not timestamp:
            timestamp = message_time
        if not timestamp:
            LOGGER.warning("Unable to get timestamp from message. Falling back to system time.")
            self._use_system_time = True
            timestamp = datetime.now(timezone.utc)

        # convert to UTC format
        timestamp = timestamp.astimezone(timezone.utc)

        # Extract register data
        data_points: List[MeterDataPoint] = []
        for obis, obj in filter(lambda o: o[1].getObjectType() == ObjectType.REGISTER, obis_obj_pairs.items()):
            reg_type = self._cosem.get_register(obis)
            if reg_type and isinstance(obj, GXDLMSRegister):
                raw_value = self._extract_register_value(obj)
                if raw_value is None:
                    LOGGER.warning("No value received for %s.", obis)
                    continue
                data_point_type = reg_type.data_point_type
                try:
                    value = float(raw_value) * reg_type.scaling
                except (TypeError, ValueError, OverflowError):
                    LOGGER.warning("Invalid register value '%s'. Skipping register.", str(raw_value))
                    continue
                data_points.append(MeterDataPoint(data_point_type, value, meter_id, timestamp))
        return data_points

    def _parse_dlms_with_push_object_list(self) -> List[GXDLMSObject]:
        parsed_objects: List[Tuple[GXDLMSObject, int]] = []
        parsed_objects = self._client.parsePushObjects(self._dlms_data.value[0])
        for index, (obj, attr_ind) in enumerate(parsed_objects[1:], start=1):
            self._client.updateValue(obj, attr_ind, self._dlms_data.value[index])
            LOGGER.debug("%s %s %s: %s", obj.objectType, obj.logicalName, attr_ind, obj.getValues()[attr_ind - 1])

        return [obj for obj, _ in parsed_objects]

    def _parse_dlms_without_push_list(self) -> List[GXDLMSObject]:
        """Tries to extract OBIS codes and values from message.
        Supported message format:
            - timestamp
            - obis code 1
            - value 1
            - obis code 2
            - value 2
            - ...
        """
        obis_codes, values = HdlcDlmsParser.extract_obis_and_values(self._dlms_data.value)
        if not obis_codes:
            return []

        push_setup = GXDLMSPushSetup()
        for obis in obis_codes:
            if obis == Cosem.CLOCK_DEFAULT_OBIS:
                push_setup.pushObjectList.append((GXDLMSClock(), GXDLMSCaptureObject(2, 0)))
            else:
                push_setup.pushObjectList.append((GXDLMSRegister(obis.to_gurux_str()), GXDLMSCaptureObject(2, 0)))

        for index, (obj, attr_ind) in enumerate(push_setup.pushObjectList):
            self._client.updateValue(obj, attr_ind.attributeIndex, values[index])
            LOGGER.debug("%d %s %s %s: %s", index, obj.objectType, obj.logicalName,
                         attr_ind.attributeIndex, obj.getValues()[attr_ind.attributeIndex - 1])
        return [obj for obj, _ in push_setup.pushObjectList]

    @staticmethod
    def _extract_value_from_data_object(data_object: GXDLMSData) -> Optional[Any]:
        return data_object.getValues()[1]

    @staticmethod
    def _extract_register_value(register: GXDLMSRegister) -> Optional[Any]:
        return register.getValues()[1]

    @staticmethod
    def extract_obis_and_values(data: List[Any]) -> Tuple[List[OBISCode], List[Any]]:
        obis_it = filter(lambda d: isinstance(d[1], (bytearray, bytes)) and OBISCode.is_obis(d[1]), enumerate(data))

        obis_codes = []
        values = []

        for idx, obis_bytes in obis_it:
            if idx + 1 < len(data) and HdlcDlmsParser.is_value(data[idx + 1]):
                obis_codes.append(OBISCode.from_bytes(obis_bytes))
                values.append(data[idx + 1])
            else:
                LOGGER.debug("Skipped OBIS code without subsequent value.")

        return obis_codes, values

    @staticmethod
    def is_value(data: Any) -> bool:
        return (
            isinstance(data, (int, str)) or
            (isinstance(data, (bytearray, bytes)) and not OBISCode.is_obis(data)))
